import { Range } from "./range";
import { RangeSetTree, Comparator } from "./range-set-tree";
import { RangeSpec } from "./range-spec";
/**
 * A set of non-intersecting ranges.
 *
 * This class provides operators for combining, subtracting, and querying
 * sets of ranges of an arbitrary type `T`. It automatically merges adjacent/intersecting
 * ranges after each operation, guaranteeing the smallest possible number of subranges
 * for every set of underlying ranges it represents.
 *
 * This class is **mutable** for performance reasons.
 *
 * This implementation is backed by a tree implementation conforming to [[`RangeSetTree`]].
 * The default implementation is `bintrees`' [[`BintreeRBRangeSetTree`]].
 *
 * @typeParam T The type of the ranges this set is expected to hold.
 */
export declare class RangeSet<T> {
    private readonly tree;
    private readonly spec;
    private constructor();
    /**
     * A shorthand constructor for numeric ranges.
     *
     * @param treeType An optional constructor for the underlying tree.
     *        Defaults to [[`BintreeRBRangeSetTree`]].
     *
     * @returns A range set built around JS `number`.
     */
    static numeric(treeType?: new (comp: Comparator<Range<number>>) => RangeSetTree<Range<number>>): RangeSet<number>;
    /**
     * A constructor method for a range set of any custom type `T`.
     *
     * @typeParam T the type of the resulting range set.
     * @param spec See [[`RangeSpec`]]
     * @param treeType An optional constructor for the underlying tree.
     *                 Defaults to [[`BintreeRBRangeSetTree`]].
     */
    static of<T>(spec: RangeSpec<T>, treeType?: new (comp: Comparator<Range<T>>) => RangeSetTree<Range<T>>): RangeSet<T>;
    /**
     * Add another range set to this set, merging intersecting/adjacent ranges.
     *
     * This operation modifies `this`, but keeps `other` untouched.
     *
     * @param other The other range set to add values from.
     *              It is **possible** to add a range set built with a different underlying tree implementation.
     */
    add(other: RangeSet<T>): this;
    /**
     * Add a range to this set, merging intersecting/adjacent ranges.
     *
     * This operation modifies `this`.
     *
     * @param other The range to add values from.
     */
    add(other: Range<T>): this;
    /**
     * Subtract every range from the provided range set from this set.
     *
     * This operation modifies `this`, but keeps `other` untouched.
     *
     * This operation relies on [[`Range.union`]].
     *
     * @param other The other range set containing the ranges to subtract.
     *              It is **possible** to subtract a range set built with a different underlying tree implementation.
     */
    subtract(other: RangeSet<T>): this;
    /**
     * Subtract `other` from this set.
     *
     * This operation modifies `this`.
     *
     * This operation relies on [[`Range.subtract`]].
     *
     * @param other The range to subtract.
     */
    subtract(other: Range<T>): this;
    /**
     * Checks whether any of the ranges within this set contains `value`.
     *
     * "Infinity" values cannot be contained, and therefore always return `false`.
     * @param value The value to check the containment of.
     */
    contains(value: T): boolean;
    /**
     * Checks whether any of the ranges within this set contains `range`.
     *
     * See [[`Range.contains`]] for details of how this check is performed.
     *
     * An empty range is always contained in any set. An empty set contains no ranges but an empty range.
     *
     * A range with an infinite bound contains other ranges with the same infinite bound, even though
     * infinity itself cannot be contained.
     *
     * @param value The value to check the containment of.
     */
    contains(range: Range<T>): boolean;
    /**
     * Finds a range containing the provided value.
     *
     * "Infinity" values cannot be contained, and therefore always return `undefined`.
     *
     * @param value The value to check the containment of.
     * @returns The range that [[`Range.contains`]] `value` if found, or `undefined` otherwise.
     */
    containing(value: T): Range<T> | undefined;
    /**
     * Finds a range enclosing the provided range.
     *
     * Enclosing means containing all the elements of `range` within the resulting range.
     *
     * Note: an empty range is **not enclosed** by any other range.
     *
     * @param range The value to be enclosed by the return.
     * @returns The range that [[`Range.contains`]] `range` if found, except an empty range, or `undefined` otherwise.
     */
    enclosing(range: Range<T>): Range<T> | undefined;
    /**
     * Clear the underlying tree, thus resetting `this` to an empty set.
     */
    clear(): this;
    /**
     * Checks whether this set is empty, i.e. whether [[`RangeSet.subranges`]] is an empty array.
     */
    get isEmpty(): boolean;
    /**
     * Returns the array of the underlying ranges of this set.
     *
     * The resulting ranges are guaranteed not to intersect with each other, nor to be adjacent to any of them.
     */
    get subranges(): Range<T>[];
}
