"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RangeSet = void 0;
const range_1 = require("./range");
const range_spec_1 = require("./range-spec");
/**
 * A set of non-intersecting ranges.
 *
 * This class provides operators for combining, subtracting, and querying
 * sets of ranges of an arbitrary type `T`. It automatically merges adjacent/intersecting
 * ranges after each operation, guaranteeing the smallest possible number of subranges
 * for every set of underlying ranges it represents.
 *
 * This class is **mutable** for performance reasons.
 *
 * This implementation is backed by a tree implementation conforming to [[`RangeSetTree`]].
 * The default implementation is `bintrees`' [[`BintreeRBRangeSetTree`]].
 *
 * @typeParam T The type of the ranges this set is expected to hold.
 */
class RangeSet {
    constructor(tree, spec) {
        this.tree = tree;
        this.spec = spec;
    }
    /**
     * A shorthand constructor for numeric ranges.
     *
     * @param treeType An optional constructor for the underlying tree.
     *        Defaults to [[`BintreeRBRangeSetTree`]].
     *
     * @returns A range set built around JS `number`.
     */
    static numeric(treeType) {
        return RangeSet.of(new range_spec_1.NumberSpec(), treeType);
    }
    /**
     * A constructor method for a range set of any custom type `T`.
     *
     * @typeParam T the type of the resulting range set.
     * @param spec See [[`RangeSpec`]]
     * @param treeType An optional constructor for the underlying tree.
     *                 Defaults to [[`BintreeRBRangeSetTree`]].
     */
    static of(spec, treeType) {
        const comp = (a, b) => {
            const upperComp = spec.comparator(a.upper, b.upper);
            if (upperComp !== 0)
                return upperComp;
            if (!a.isUpperEnclosed && b.isUpperEnclosed)
                return -1;
            if (a.isUpperEnclosed && !b.isUpperEnclosed)
                return 1;
            return 0;
        };
        if (!treeType) {
            const defaultAdapter = require("./range-set-tree-bintree");
            treeType = defaultAdapter.BintreeRBRangeSetTree;
            if (!treeType) {
                throw Error(`No compatible binary tree implementation was provided for this RangeSet: ` +
                    `you should either install "bintree", ` +
                    `or provide a compatible binary tree adapter by implementing RangeSetTree and supplying its constructor to this of() invocation`);
            }
        }
        return new RangeSet(new treeType(comp), spec);
    }
    add(other) {
        if (other instanceof RangeSet) {
            other.subranges.forEach((r) => this.add(r));
            return this;
        }
        if (other.isEmpty)
            return this;
        if (this.isEmpty) {
            this.tree.insert(other);
            return this;
        }
        const foundIt = this.tree.upperBound(other);
        if (!foundIt.data) {
            foundIt.prev();
        }
        let newRange = other;
        const toDelete = [];
        for (let item = foundIt.data; item !== null && this.spec.isGreaterOrEqualTo(item.upper, newRange.lower); item = foundIt.prev()) {
            if (!newRange.intersection(item).isEmpty || newRange.adjacent(item)) {
                toDelete.push(item);
                newRange = newRange.union(item);
            }
        }
        toDelete.forEach((r) => this.tree.remove(r));
        if (!newRange.isEmpty)
            this.tree.insert(newRange);
        return this;
    }
    subtract(other) {
        if (other instanceof RangeSet) {
            other.subranges.forEach((r) => this.subtract(r));
            return this;
        }
        if (this.isEmpty) {
            return this;
        }
        const foundIt = this.tree.upperBound(other);
        if (!foundIt.data) {
            foundIt.prev();
        }
        const toAdd = [];
        const toDelete = [];
        for (let item = foundIt.data; item !== null && this.spec.isGreaterOrEqualTo(item.upper, other.lower); item = foundIt.prev()) {
            if (other.intersection(item).isEmpty)
                continue;
            toDelete.push(item);
            item.subtract(other).forEach((r) => toAdd.push(r));
        }
        toDelete.forEach((r) => this.tree.remove(r));
        toAdd.filter((r) => !r.isEmpty).forEach((r) => this.tree.insert(r));
        return this;
    }
    contains(rangeOrValue) {
        if (rangeOrValue instanceof range_1.Range) {
            return rangeOrValue.isEmpty || !!this.enclosing(rangeOrValue);
        }
        else {
            return !!this.containing(rangeOrValue);
        }
    }
    /**
     * Finds a range containing the provided value.
     *
     * "Infinity" values cannot be contained, and therefore always return `undefined`.
     *
     * @param value The value to check the containment of.
     * @returns The range that [[`Range.contains`]] `value` if found, or `undefined` otherwise.
     */
    containing(value) {
        if (this.spec.isInfinity(value))
            return; // It cannot contain infinity
        const singleton = range_1.Range.singleton(value, this.spec);
        return this.enclosing(singleton);
    }
    /**
     * Finds a range enclosing the provided range.
     *
     * Enclosing means containing all the elements of `range` within the resulting range.
     *
     * Note: an empty range is **not enclosed** by any other range.
     *
     * @param range The value to be enclosed by the return.
     * @returns The range that [[`Range.contains`]] `range` if found, except an empty range, or `undefined` otherwise.
     */
    enclosing(range) {
        if (range.isEmpty)
            return; // No range encloses an empty range
        const foundIt = this.tree.lowerBound(range);
        const candidateRange = foundIt.data;
        if (!candidateRange || !candidateRange.contains(range))
            return;
        return candidateRange;
    }
    /**
     * Clear the underlying tree, thus resetting `this` to an empty set.
     */
    clear() {
        this.tree.clear();
        return this;
    }
    /**
     * Checks whether this set is empty, i.e. whether [[`RangeSet.subranges`]] is an empty array.
     */
    get isEmpty() {
        return this.tree.size === 0;
    }
    /**
     * Returns the array of the underlying ranges of this set.
     *
     * The resulting ranges are guaranteed not to intersect with each other, nor to be adjacent to any of them.
     */
    get subranges() {
        const result = [];
        const it = this.tree.iterator();
        let value;
        while ((value = it.prev()) !== null) {
            result.push(value);
        }
        return result.reverse();
    }
}
exports.RangeSet = RangeSet;
